//字符串的模式匹配算法的验证
//要点：
   //1. 函数nfind(char * const str, char * const pat)在str中查找模式串pat。
   //2. 查找的方法是顺序比较模式串中的每个字符和主串的每个字符，直到发现匹配或扫描到主串的结束。
   //3. 为提高效率，1）发现模式串的长度大于主串剩下的串长就结束。
   //   2）如果pat的第一个字符和str中的字符匹配，则比较pat的最后一个字符是否和str对应位置的字符匹配。
   //4. 算法及程序实现参见《Fundamentals of DATA STRUCTURES IN C》
//关键字：字符串、内置谓词、逻辑变量

#include <limits.h>
#include <string.h>

//////////////////////////////////////////逻辑变量/////////////////////////////////////////////////
//@ logic integer m, n;         //逻辑变量m和n分别记录物理字符串str和pat的长度

/////////////////////////////////////////nfind函数的验证////////////////////////////////////////////
//nfind的函数协议：
//前条件：描述函数入口处的两个物理字符串
//后条件：
//函数返回值等于-1，则主串和模式串不匹配
//函数返回值是非负数，则主串和模式串匹配，且该返回值是匹配开始位置 

/*@ 
  requires 
      \is_pstring(str, m) && \is_pstring(pat, n) 
      && m > 0 && n > 0 && m + n <= INT_MAX
      ;
  ensures 
      \is_pstring(str, m) && \is_pstring(pat, n) 
      && 
      (
        \result == -1 && !\contains(\string(str, m), \string(pat, n))
      ||
        0 <= \result < m && (\forall integer i:[0..n-1]. str[\result + i] == pat[i])
        && (\forall integer i:[0..\result - 1]. \exists integer j:[0..n-1].str[i+j] != pat[j])
      )
      ;
*/    

int nfind(char * const str, char * const pat)
{
    int i,j=-1,start = 0;
    int lasts = strlen(str) - 1;
    int lastp = strlen(pat) - 1;
    int endmatch = lastp;
    
    /*@ loop invariant
          \is_pstring(str, m) && \is_pstring(pat, n) 
          && lasts == m - 1 && lastp == n - 1 && endmatch >= lastp 
          && endmatch - start == n - 1 && start >= 0 && m > 0 && n > 0 && m + n <= INT_MAX
          && (\forall integer i:[0..start-1]. \exists integer j:[0..n-1].str[i+j] != pat[j]) 
          && j < lastp
          ;      
    */
    for(i = 0; endmatch <= lasts; endmatch++,start++){
        if(str[endmatch] == pat[lastp])
            /*@ loop invariant
                  \is_pstring(str, m) && \is_pstring(pat, n) 
                  && lasts == m - 1 && lastp == n - 1 && endmatch >= lastp 
                  && endmatch <= lasts && start >= 0 
                  && str[endmatch] == pat[lastp] && m > 0 && n > 0 && m + n <= INT_MAX
                  && 0 <= j <= lastp && i >= start && endmatch - start == n - 1 
                  && (\forall integer i:[0..j-1]. str[start + i] == pat[i])
                  && (\forall integer i:[0..start-1].\exists integer j:[0..n-1].str[i+j] != pat[j]) 
                  && i - start == j 
                  ;
            */
            for(j = 0,i = start; j<lastp && str[i] == pat[j]; i++,j++)
                ;
        if (j == lastp) 
            return start;
    }
    return -1;
}